@@ -1,6 +1,6 @@
 ;; Continuation-passing style (CPS) intermediate language (IL)
 
-;; Copyright (C) 2013-2020 Free Software Foundation, Inc.
+;; Copyright (C) 2013-2021 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -30,6 +30,7 @@
   #:use-module (srfi srfi-11)
   #:use-module (srfi srfi-26)
   #:use-module (language cps)
+  #:use-module (language cps graphs)
   #:use-module (language cps utils)
   #:use-module (language cps intmap)
   #:use-module (language cps intset)
@@ -39,7 +40,8 @@
             lookup-representation
             lookup-nlocals
             lookup-call-proc-slot
-            lookup-parallel-moves
+            lookup-send-parallel-moves
+            lookup-receive-parallel-moves
             lookup-slot-map))
 
 (define-record-type $allocation
@@ -56,8 +58,8 @@
   ;;
   (representations allocation-representations)
 
-  ;; A map of LABEL to /call allocs/, for expressions that continue to
-  ;; $kreceive continuations: non-tail calls and $prompt terms.
+  ;; A map of LABEL to /call allocs/, for non-tail $call/$callk, and for
+  ;; $prompt.
   ;;
   ;; A call alloc contains two pieces of information: the call's /proc
   ;; slot/ and a /dead slot map/.  The proc slot indicates the slot of a
@@ -72,7 +74,7 @@
 
   ;; A map of LABEL to /parallel moves/.  Parallel moves shuffle locals
   ;; into position for a $call, $callk, or $values, or shuffle returned
-  ;; values back into place in a $kreceive.
+  ;; values back into place at a return continuation.
   ;;
   ;; A set of moves is expressed as an ordered list of (SRC . DST)
   ;; moves, where SRC and DST are slots.  This may involve a temporary
@@ -111,8 +113,13 @@
   (or (call-alloc-proc-slot (lookup-call-alloc k allocation))
       (error "Call has no proc slot" k)))
 
-(define (lookup-parallel-moves k allocation)
-  (intmap-ref (allocation-shuffles allocation) k))
+(define (lookup-send-parallel-moves k allocation)
+  (match (intmap-ref (allocation-shuffles allocation) k)
+    ((send . receive) send)))
+
+(define (lookup-receive-parallel-moves k allocation)
+  (match (intmap-ref (allocation-shuffles allocation) k)
+    ((send . receive) receive)))
 
 (define (lookup-slot-map k allocation)
   (or (call-alloc-slot-map (lookup-call-alloc k allocation))
@@ -121,89 +128,6 @@
 (define (lookup-nlocals allocation)
   (allocation-frame-size allocation))
 
-(define-syntax-rule (persistent-intmap2 exp)
-  (call-with-values (lambda () exp)
-    (lambda (a b)
-      (values (persistent-intmap a) (persistent-intmap b)))))
-
-(define (compute-defs-and-uses cps)
-  "Return two LABEL->VAR... maps indicating values defined at and used
-by a label, respectively."
-  (define (vars->intset vars)
-    (fold (lambda (var set) (intset-add set var)) empty-intset vars))
-  (persistent-intmap2
-   (intmap-fold
-    (lambda (label cont defs uses)
-      (define (get-defs k)
-        (match (intmap-ref cps k)
-          (($ $kargs names vars) (vars->intset vars))
-          (_ empty-intset)))
-      (define (return d u)
-        (values (intmap-add! defs label d)
-                (intmap-add! uses label u)))
-      (match cont
-        (($ $kfun src meta self)
-         (return (if self (intset self) empty-intset) empty-intset))
-        (($ $kargs _ _ ($ $continue k src exp))
-         (match exp
-           ((or ($ $const) ($ $const-fun) ($ $code))
-            (return (get-defs k) empty-intset))
-           (($ $call proc args)
-            (return (get-defs k) (intset-add (vars->intset args) proc)))
-           (($ $callk _ proc args)
-            (let ((args (vars->intset args)))
-              (return (get-defs k) (if proc (intset-add args proc) args))))
-           (($ $primcall name param args)
-            (return (get-defs k) (vars->intset args)))
-           (($ $values args)
-            (return (get-defs k) (vars->intset args)))))
-        (($ $kargs _ _ ($ $branch kf kt src op param args))
-         (return empty-intset (vars->intset args)))
-        (($ $kargs _ _ ($ $prompt k kh src escape? tag))
-         (return empty-intset (intset tag)))
-        (($ $kargs _ _ ($ $throw src op param args))
-         (return empty-intset (vars->intset args)))
-        (($ $kclause arity body alt)
-         (return (get-defs body) empty-intset))
-        (($ $kreceive arity kargs)
-         (return (get-defs kargs) empty-intset))
-        (($ $ktail)
-         (return empty-intset empty-intset))))
-    cps
-    empty-intmap
-    empty-intmap)))
-
-(define (compute-reverse-control-flow-order preds)
-  "Return a LABEL->ORDER bijection where ORDER is a contiguous set of
-integers starting from 0 and incrementing in sort order.  There is a
-precondition that labels in PREDS are already renumbered in reverse post
-order."
-  (define (has-back-edge? preds)
-    (let/ec return
-      (intmap-fold (lambda (label labels)
-                     (intset-fold (lambda (pred)
-                                    (if (<= label pred)
-                                        (return #t)
-                                        (values)))
-                                  labels)
-                     (values))
-                   preds)
-      #f))
-  (if (has-back-edge? preds)
-      ;; This is more involved than forward control flow because not all
-      ;; live labels are reachable from the tail.
-      (persistent-intmap
-       (fold2 (lambda (component order n)
-                (intset-fold (lambda (label order n)
-                               (values (intmap-add! order label n)
-                                       (1+ n)))
-                             component order n))
-              (reverse (compute-sorted-strongly-connected-components preds))
-              empty-intmap 0))
-      ;; Just reverse forward control flow.
-      (let ((max (intmap-prev preds)))
-        (intmap-map (lambda (label labels) (- max label)) preds))))
-
 (define* (add-prompt-control-flow-edges conts succs #:key complete?)
   "For all prompts in DFG in the range [MIN-LABEL, MIN-LABEL +
 LABEL-COUNT), invoke F with arguments PROMPT, HANDLER, and BODY for each
@@ -236,6 +160,10 @@ body continuation in the prompt."
               (visit-cont k level labels))
              (($ $kargs names syms ($ $branch kf kt))
               (visit-cont kf level (visit-cont kt level labels)))
+             (($ $kargs names syms ($ $switch kf kt*))
+              (fold1 (lambda (label labels)
+                       (visit-cont label level labels))
+                     (cons kf kt*) labels))
              (($ $kargs names syms ($ $prompt k kh src escape? tag))
               (visit-cont kh level (visit-cont k (1+ level) labels)))
              (($ $kargs names syms ($ $throw)) labels))))))))
@@ -263,51 +191,6 @@ body continuation in the prompt."
    conts
    succs))
 
-(define (rename-keys map old->new)
-  (persistent-intmap
-   (intmap-fold (lambda (k v out)
-                  (intmap-add! out (intmap-ref old->new k) v))
-                map
-                empty-intmap)))
-
-(define (rename-intset set old->new)
-  (intset-fold (lambda (old set) (intset-add set (intmap-ref old->new old)))
-               set empty-intset))
-
-(define (rename-graph graph old->new)
-  (persistent-intmap
-   (intmap-fold (lambda (pred succs out)
-                  (intmap-add! out
-                               (intmap-ref old->new pred)
-                               (rename-intset succs old->new)))
-                graph
-                empty-intmap)))
-
-(define (compute-live-variables cps defs uses)
-  "Compute and return two values mapping LABEL->VAR..., where VAR... are
-the definitions that are live before and after LABEL, as intsets."
-  (let* ((succs (add-prompt-control-flow-edges cps (compute-successors cps)))
-         (preds (invert-graph succs))
-         (old->new (compute-reverse-control-flow-order preds))
-         (init (persistent-intmap (intmap-fold
-                                   (lambda (old new init)
-                                     (intmap-add! init new empty-intset))
-                                   old->new empty-intmap))))
-    (call-with-values
-        (lambda ()
-          (solve-flow-equations (rename-graph preds old->new)
-                                init init
-                                (rename-keys defs old->new)
-                                (rename-keys uses old->new)
-                                intset-subtract intset-union intset-union))
-      (lambda (in out)
-        ;; As a reverse control-flow problem, the values flowing into a
-        ;; node are actually the live values after the node executes.
-        ;; Funny, innit?  So we return them in the reverse order.
-        (let ((new->old (invert-bijection old->new)))
-          (values (rename-keys out new->old)
-                  (rename-keys in new->old)))))))
-
 (define (compute-needs-slot cps defs uses)
   (define (get-defs k) (intmap-ref defs k))
   (define (get-uses label) (intmap-ref uses label))
@@ -325,7 +208,7 @@ the definitions that are live before and after LABEL, as intsets."
         (($ $kclause arity body alternate)
          (get-defs label))
         (($ $kfun src meta self)
-         (if self (intset self) empty-intset))
+         (get-defs label))
         (($ $ktail)
          empty-intset))))
    cps
@@ -533,18 +416,28 @@ are comparable with eqv?.  A tmp slot may be used."
   (define (parallel-move src-slots dst-slots tmp-slot)
     (solve-parallel-move src-slots dst-slots tmp-slot))
 
-  (define (compute-receive-shuffles label proc-slot)
-    (match (get-cont label)
+  ;; A term can have two sets of shuffles: one set to shuffle operands
+  ;; to the term (the "send moves"), and one set to shuffle results (the
+  ;; "receive moves").  An example of send moves would be a call getting
+  ;; its arguments into position, or a $values performing a parallel
+  ;; move.  Receive moves come when binding call results to values, for
+  ;; local returns (call returns) or non-local returns (prompt
+  ;; handlers).
+  (define (add-shuffles shuffles label send-moves receive-moves)
+    (intmap-add! shuffles label (cons send-moves receive-moves)))
+
+  (define (compute-receive-shuffles k proc-slot)
+    (match (get-cont k)
       (($ $kreceive arity kargs)
-       (let* ((results (match (get-cont kargs)
-                         (($ $kargs names vars) vars)))
-              (value-slots (integers proc-slot (length results)))
+       (compute-receive-shuffles kargs proc-slot))
+      (($ $kargs names results)
+       (let* ((value-slots (integers proc-slot (length results)))
               (result-slots (get-slots results))
               ;; Filter out unused results.
               (value-slots (filter-map (lambda (val result) (and result val))
                                        value-slots result-slots))
               (result-slots (filter (lambda (x) x) result-slots))
-              (live (compute-live-slots kargs)))
+              (live (compute-live-slots k)))
          (parallel-move value-slots
                         result-slots
                         (compute-tmp-slot live value-slots))))))
@@ -554,19 +447,19 @@ are comparable with eqv?.  A tmp slot may be used."
       (($ $ktail)
        (let* ((live (compute-live-slots label))
               (tail-slots (integers 0 (length args)))
-              (moves (parallel-move (get-slots args)
-                                    tail-slots
-                                    (compute-tmp-slot live tail-slots))))
-         (intmap-add! shuffles label moves)))
-      (($ $kreceive)
+              (send-moves (parallel-move (get-slots args)
+                                         tail-slots
+                                         (compute-tmp-slot live tail-slots))))
+         (add-shuffles shuffles label send-moves '())))
+      ((or ($ $kargs) ($ $kreceive))
        (let* ((live (compute-live-slots label))
               (proc-slot (get-proc-slot label))
               (call-slots (integers proc-slot (length args)))
-              (arg-moves (parallel-move (get-slots args)
-                                        call-slots
-                                        (compute-tmp-slot live call-slots))))
-         (intmap-add! (intmap-add! shuffles label arg-moves)
-                      k (compute-receive-shuffles k proc-slot))))))
+              (send-moves (parallel-move (get-slots args)
+                                         call-slots
+                                         (compute-tmp-slot live call-slots)))
+              (receive-moves (compute-receive-shuffles k proc-slot)))
+         (add-shuffles shuffles label send-moves receive-moves)))))
     
   (define (add-values-shuffles label k args shuffles)
     (match (get-cont k)
@@ -574,21 +467,22 @@ are comparable with eqv?.  A tmp slot may be used."
        (let* ((live (compute-live-slots label))
               (src-slots (get-slots args))
               (dst-slots (integers 0 (length args)))
-              (moves (parallel-move src-slots dst-slots
-                                    (compute-tmp-slot live dst-slots))))
-         (intmap-add! shuffles label moves)))
+              (send-moves (parallel-move src-slots dst-slots
+                                         (compute-tmp-slot live dst-slots))))
+         (add-shuffles shuffles label send-moves '())))
       (($ $kargs _ dst-vars)
        (let* ((live (logior (compute-live-slots label)
                             (compute-live-slots k)))
               (src-slots (get-slots args))
               (dst-slots (get-slots dst-vars))
-              (moves (parallel-move src-slots dst-slots
-                                    (compute-tmp-slot live '()))))
-         (intmap-add! shuffles label moves)))))
+              (send-moves (parallel-move src-slots dst-slots
+                                         (compute-tmp-slot live '()))))
+         (add-shuffles shuffles label send-moves '())))))
 
   (define (add-prompt-shuffles label k handler shuffles)
-    (intmap-add! shuffles handler
-                 (compute-receive-shuffles handler (get-proc-slot label))))
+    (define receive-moves
+      (compute-receive-shuffles handler (get-proc-slot label)))
+    (add-shuffles shuffles label '() receive-moves))
 
   (define (compute-shuffles label cont shuffles)
     (match cont
@@ -623,11 +517,14 @@ are comparable with eqv?.  A tmp slot may be used."
       (slot (max size (1+ slot)))))
   (define (max-size* vars size)
     (fold max-size size vars))
-  (define (shuffle-size moves size)
+  (define (shuffle-size* moves size)
     (match moves
       (() size)
       (((src . dst) . moves)
-       (shuffle-size moves (max size (1+ src) (1+ dst))))))
+       (shuffle-size* moves (max size (1+ src) (1+ dst))))))
+  (define (shuffle-size send+receive size)
+    (match send+receive
+      ((send . receive) (shuffle-size* send (shuffle-size* receive size)))))
   (define (call-size label nargs size)
     (shuffle-size (get-shuffles label)
                   (max (+ (get-proc-slot label) nargs) size)))
@@ -643,35 +540,37 @@ are comparable with eqv?.  A tmp slot may be used."
               (call-size label (+ nclosure (length args)) size)))
            (($ $continue _ _ ($ $values args))
             (shuffle-size (get-shuffles label) size))
+           (($ $prompt)
+            (shuffle-size (get-shuffles label) size))
            (_ size))))
-      (($ $kreceive)
-       (shuffle-size (get-shuffles label) size))
       (_ size)))
 
   (intmap-fold measure-cont cps minimum-frame-size))
 
 (define (allocate-args cps)
+  (define (add-clause entry first-slot slots)
+    (match (intmap-ref cps entry)
+      (($ $kclause arity body alt)
+       (let ((slots (add-clause body first-slot slots)))
+         (if alt
+             (add-clause alt first-slot slots)
+             slots)))
+      (($ $kargs names vars)
+       (let lp ((vars vars) (n first-slot) (slots slots))
+         (match vars
+           (() slots)
+           ((var . vars)
+            (lp vars
+                (1+ n)
+                (intmap-add slots var n))))))))
   (match (intmap-ref cps (intmap-next cps))
-    (($ $kfun _ _ has-self?)
-     (intmap-fold (lambda (label cont slots)
-                    (match cont
-                      (($ $kfun src meta self)
-                       (if has-self?
-                           (intmap-add! slots self 0)
-                           slots))
-                      (($ $kclause arity body alt)
-                       (match (intmap-ref cps body)
-                         (($ $kargs names vars)
-                          (let lp ((vars vars) (slots slots)
-                                   (n (if has-self? 1 0)))
-                            (match vars
-                              (() slots)
-                              ((var . vars)
-                               (lp vars
-                                   (intmap-add! slots var n)
-                                   (1+ n))))))))
-                      (_ slots)))
-                  cps empty-intmap))))
+    (($ $kfun src meta self tail entry)
+     (add-clause
+      entry
+      (if self 1 0)
+      (if self
+          (intmap-add empty-intmap self 0)
+          empty-intmap)))))
 
 (define (allocate-lazy-vars cps slots call-allocs live-in lazy)
   (define (compute-live-slots slots label)
@@ -735,81 +634,14 @@ are comparable with eqv?.  A tmp slot may be used."
   (persistent-intmap
    (intmap-fold-right allocate-lazy cps slots)))
 
-(define (compute-var-representations cps)
-  (define (get-defs k)
-    (match (intmap-ref cps k)
-      (($ $kargs names vars) vars)
-      (_ '())))
-  (intmap-fold
-   (lambda (label cont representations)
-     (match cont
-       (($ $kargs _ _ ($ $continue k _ exp))
-        (match (get-defs k)
-          (() representations)
-          ((var)
-           (match exp
-             (($ $values (arg))
-              (intmap-add representations var
-                          (intmap-ref representations arg)))
-             (($ $primcall (or 'scm->f64 'load-f64 's64->f64
-                               'f32-ref 'f64-ref
-                               'fadd 'fsub 'fmul 'fdiv 'fsqrt 'fabs
-                               'ffloor 'fceiling
-                               'fsin 'fcos 'ftan 'fasin 'facos 'fatan 'fatan2))
-              (intmap-add representations var 'f64))
-             (($ $primcall (or 'scm->u64 'scm->u64/truncate 'load-u64
-                               's64->u64
-                               'assume-u64
-                               'uadd 'usub 'umul
-                               'ulogand 'ulogior 'ulogxor 'ulogsub 'ursh 'ulsh
-                               'uadd/immediate 'usub/immediate 'umul/immediate
-                               'ursh/immediate 'ulsh/immediate
-                               'u8-ref 'u16-ref 'u32-ref 'u64-ref
-                               'word-ref 'word-ref/immediate
-                               'untag-char))
-              (intmap-add representations var 'u64))
-             (($ $primcall (or 'untag-fixnum
-                               'assume-s64
-                               'scm->s64 'load-s64 'u64->s64
-                               'srsh 'srsh/immediate
-                               's8-ref 's16-ref 's32-ref 's64-ref))
-              (intmap-add representations var 's64))
-             (($ $primcall (or 'pointer-ref/immediate
-                               'tail-pointer-ref/immediate))
-              (intmap-add representations var 'ptr))
-             (($ $code)
-              (intmap-add representations var 'u64))
-             (_
-              (intmap-add representations var 'scm))))
-          (vars
-           (match exp
-             (($ $values args)
-              (fold (lambda (arg var representations)
-                      (intmap-add representations var
-                                  (intmap-ref representations arg)))
-                    representations args vars))))))
-       (($ $kargs _ _ (or ($ $branch) ($ $prompt) ($ $throw)))
-        representations)
-       (($ $kfun src meta self)
-        (if self
-            (intmap-add representations self 'scm)
-            representations))
-       (($ $kclause arity body alt)
-        (fold1 (lambda (var representations)
-                 (intmap-add representations var 'scm))
-               (get-defs body) representations))
-       (($ $kreceive arity kargs)
-        (fold1 (lambda (var representations)
-                 (intmap-add representations var 'scm))
-               (get-defs kargs) representations))
-       (($ $ktail) representations)))
-   cps
-   empty-intmap))
-
 (define* (allocate-slots cps #:key (precolor-calls? #t))
   (let*-values (((defs uses) (compute-defs-and-uses cps))
                 ((representations) (compute-var-representations cps))
-                ((live-in live-out) (compute-live-variables cps defs uses))
+                ((live-in live-out)
+                 (let* ((succs (compute-successors cps))
+                        (succs+ (add-prompt-control-flow-edges cps succs))
+                        (preds (invert-graph succs+)))
+                   (compute-live-variables preds defs uses)))
                 ((needs-slot) (compute-needs-slot cps defs uses))
                 ((lazy) (if precolor-calls?
                             (compute-lazy-vars cps live-in live-out defs
@@ -917,6 +749,8 @@ are comparable with eqv?.  A tmp slot may be used."
            (values (allocate* args tail-slots slots pre-live)
                    call-allocs)))
         (($ $kreceive arity kargs)
+         (allocate-call label kargs args slots call-allocs pre-live))
+        (($ $kargs names results)
          (let*-values
              (((post-live) (compute-live-out-slots slots label))
               ((proc-slot) (compute-call-proc-slot post-live))
@@ -928,13 +762,13 @@ are comparable with eqv?.  A tmp slot may be used."
               ;; especially for unused extra values, and avoiding frame
               ;; size growth due to sparse locals.
               ((slots result-live)
-               (match (get-cont kargs)
-                 (($ $kargs () ())
+               (match results
+                 (()
                   (values slots post-live))
-                 (($ $kargs (_ . _) (_ . results))
+                 ((_ . results*)
                   (let ((result-slots (integers (+ proc-slot 1)
-                                                (length results))))
-                    (allocate* results result-slots slots post-live)))))
+                                                (length results*))))
+                    (allocate* results* result-slots slots post-live)))))
               ((slot-map) (compute-slot-map slots (intmap-ref live-out label)
                                             (- proc-slot frame-size)))
               ((call) (make-call-alloc proc-slot slot-map)))
