@@ -1,6 +1,6 @@
 ;;; installed-scm-file
 
-;;;; Copyright (C) 2000,2001,2002, 2006, 2009, 2010 Free Software Foundation, Inc.
+;;;; Copyright (C) 2000,2001,2002, 2006, 2009, 2010, 2013, 2015, 2020 Free Software Foundation, Inc.
 ;;;; 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -19,13 +19,15 @@
 
 
 (define-module (oop goops save)
-  :use-module (oop goops internal)
-  :use-module (oop goops util)
-  :re-export (make-unbound)
-  :export (save-objects load-objects restore
-	   enumerate! enumerate-component!
-	   write-readably write-component write-component-procedure
-	   literal? readable make-readable))
+  #:use-module (ice-9 copy-tree)
+  #:use-module (oop goops internal)
+  #:export (make-unbound save-objects load-objects restore
+	    enumerate! enumerate-component!
+	    write-readably write-component write-component-procedure
+	    literal? readable make-readable))
+
+(define (make-unbound)
+  *unbound*)
 
 ;;;
 ;;; save-objects ALIST PORT [EXCLUDED] [USES]
@@ -122,8 +124,33 @@
 (define (readable-expression obj)
   `(readable ,(hashq-ref readables obj)))
 
+;; FIXME: if obj is nil or false, this can return a false value.  OTOH
+;; usually this is only for non-immediates.
 (define (readable? obj)
-  (hashq-get-handle readables obj))
+  (hashq-ref readables obj))
+
+;;;
+;;; Writer helpers
+;;;
+
+(define (write-component-procedure o file env)
+  "Return #f if circular reference"
+  (cond ((immediate? o) (write o file) #t)
+	((readable? o) (write (readable-expression o) file) #t)
+	((excluded? o env) (display #f file) #t)
+	(else
+	 (let ((info (object-info o env)))
+	   (cond ((not (binding? info)) (write-readably o file env) #t)
+		 ((not (eq? (visiting info) #:defined)) #f) ;forward reference
+		 (else (display (binding info) file) #t))))))
+
+;;; write-component OBJECT PATCHER FILE ENV
+;;;
+(define-macro (write-component object patcher file env)
+  `(or (write-component-procedure ,object ,file ,env)
+       (begin
+         (display #f ,file)
+         (add-patcher! ,patcher ,env))))
 
 ;;;
 ;;; Strings
@@ -344,33 +371,32 @@
 ;; Don't export this function!  This is all very temporary.
 ;;
 (define (get-set-for-each proc class)
-  (for-each (lambda (slotdef g-n-s)
-	      (let ((g-n-s (cddr g-n-s)))
-		(cond ((integer? g-n-s)
-		       (proc (standard-get g-n-s) (standard-set g-n-s)))
-		      ((not (memq (slot-definition-allocation slotdef)
-				  '(#:class #:each-subclass)))
-		       (proc (car g-n-s) (cadr g-n-s))))))
-	    (class-slots class)
-	    (slot-ref class 'getters-n-setters)))
+  (for-each (lambda (slot)
+	      (unless (memq (slot-definition-allocation slot)
+                            '(#:class #:each-subclass))
+                (let ((ref (slot-definition-slot-ref slot))
+                      (set (slot-definition-slot-set! slot))
+                      (index (slot-definition-index slot)))
+                  (if ref
+                      (proc ref set)
+                      (proc (standard-get index) (standard-set index))))))
+	    (class-slots class)))
 
 (define (access-for-each proc class)
-  (for-each (lambda (slotdef g-n-s)
-	      (let ((g-n-s (cddr g-n-s))
-		    (a (slot-definition-accessor slotdef)))
-		(cond ((integer? g-n-s)
-		       (proc (slot-definition-name slotdef)
-			     (and a (generic-function-name a))
-			     (standard-get g-n-s)
-			     (standard-set g-n-s)))
-		      ((not (memq (slot-definition-allocation slotdef)
-				  '(#:class #:each-subclass)))
-		       (proc (slot-definition-name slotdef)
-			     (and a (generic-function-name a))
-			     (car g-n-s)
-			     (cadr g-n-s))))))
-	    (class-slots class)
-	    (slot-ref class 'getters-n-setters)))
+  (for-each (lambda (slot)
+              (unless (memq (slot-definition-allocation slot)
+                            '(#:class #:each-subclass))
+                (let ((name (slot-definition-name slot))
+                      (accessor (and=> (slot-definition-accessor slot)
+                                       generic-function-name))
+                      (ref (slot-definition-slot-ref slot))
+                      (set (slot-definition-slot-set! slot))
+                      (index (slot-definition-index slot)))
+                  (if ref
+                      (proc name accessor ref set)
+                      (proc name accessor
+                            (standard-get index) (standard-set index))))))
+	    (class-slots class)))
 
 (define-macro (restore class slots . exps)
   "(restore CLASS (SLOT-NAME1 ...) EXP1 ...)"
@@ -601,24 +627,6 @@
 	   (pop-ref! env)
 	   (set! (objects env) (cons o (objects env)))))))
 
-(define (write-component-procedure o file env)
-  "Return #f if circular reference"
-  (cond ((immediate? o) (write o file) #t)
-	((readable? o) (write (readable-expression o) file) #t)
-	((excluded? o env) (display #f file) #t)
-	(else
-	 (let ((info (object-info o env)))
-	   (cond ((not (binding? info)) (write-readably o file env) #t)
-		 ((not (eq? (visiting info) #:defined)) #f) ;forward reference
-		 (else (display (binding info) file) #t))))))
-
-;;; write-component OBJECT PATCHER FILE ENV
-;;;
-(define-macro (write-component object patcher file env)
-  `(or (write-component-procedure ,object ,file ,env)
-       (begin
-         (display #f ,file)
-         (add-patcher! ,patcher ,env))))
 
 ;;;
 ;;; Main engine
